package src.desktop;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import src.VisualXML;

/**
 * Class provides scaled view on the desktop and all of its Internal Frames. 
 * DesktopView contains an inner class MovableView, which simulates current 
 * Viewport of the desktop. 
 */  
public class DesktopView extends JPanel
{
   protected VisualXML parent; // parent main window
   protected int scale = 5;    // scale coefficient
 
   /**
    * Constructor. Creates and displays mini frames and movable rectangle
    *     
    * @param f parent window 
    */     
   public DesktopView(JFrame f)
   {
      super();
      setBackground(Color.white);
      setLayout(null);
      parent = (VisualXML) f;
      
      // create and display content
      showView();
      drawMiniFrames();
   }

   /**
    * Sets the scale coefficient of the view
    * 
    * @param s new scale        
    */       
   public void setScale(int s)
   {
      scale = s;
   }
   
   /**
    * Returns current scale coefficient of the view
    */    
   public int getScale()
   {
      return scale;
   }
   
   /**
    * Creates and displays movable rectangle (which simulates current view)
    */       
   protected void showView()
   {
      MovableView mv = new MovableView();
      Rectangle bounds = ((JViewport)parent.desktop.getParent()).getViewRect();
      mv.setBounds(bounds.x/scale+5, bounds.y/scale+5, bounds.width/scale, bounds.height/scale);
      add(mv);
   }
   
   /**
    * Returns preferred size
    */       
   public Dimension getPreferredSize()
   {
      return new Dimension(parent.desktop.getWidth()/scale+10,parent.desktop.getHeight()/scale+10);
   }
   
   /**
    * Refreshes content. (Recreates all components)
    */       
   public void refresh()
   {
      removeAll();
      showView();
      drawMiniFrames();
      
   }
   
   /**
    * Creates all mini frames
    */       
   public void drawMiniFrames()
   {
      JInternalFrame allFrames[] = parent.desktop.getAllFrames();

      for (int i = 0; i < allFrames.length; i++)
      {// for all iframes
         if (allFrames[i] instanceof InnerFrame)
         {// for InnerFrames only
            createMiniFrame(allFrames[i].getNormalBounds());
         }
      }
   }
   
   /**
    * Creates scaled mini frame
    * 
    * @param r Bounds of original iframe.        
    */       
   protected void createMiniFrame(Rectangle r)
   {
      JPanel p = new JPanel();
      p.setBackground(new Color(192,192,192));
      p.setBounds(r.x/scale+5, r.y/scale+5, r.width/scale, r.height/scale);
      add(p);
   }
   
   /**
    * Represents "movable rectangle"
    */       
   class MovableView extends JComponent implements MouseInputListener
   {
      int m_XDifference, m_YDifference;
      protected int width;
      protected int height;
      boolean m_dragging;
      
      /**
       * Constructor
       */             
      public MovableView()
      {
         setBorder(BorderFactory.createLineBorder(Color.black));
         addMouseListener(this);
         addMouseMotionListener(this);
      }
      
      // needed for MouseInputListener interface
      public void mouseClicked(MouseEvent e){}
      public void mouseEntered(MouseEvent e){}
      public void mouseExited(MouseEvent e){}
      public void mouseMoved(MouseEvent e){}
      
      /**
       * Sets new location of rectangle and of the Viewport
       */             
      public void mouseDragged(MouseEvent e)
      {
         if (m_dragging)
         {
            JViewport vp = (JViewport)parent.desktop.getParent();
            int x = e.getX()-m_XDifference + getX();
            int y = e.getY()-m_YDifference + getY();
            setLocation(x, y);
         
            if (x < 0) x = 0;
            if (y < 0) y = 0;
         
            if (x*scale+vp.getWidth() > parent.desktop.getWidth())
            {
               x = parent.desktop.getWidth() - vp.getWidth();
            }
            
            if (y*scale+vp.getHeight() > parent.desktop.getHeight())
            {
               y = parent.desktop.getHeight() - vp.getHeight();
            }
         
            vp.setViewPosition(new Point(x*scale,y*scale));
         }
      }

      /**
       * Start dragging. 
       */             
      public void mousePressed(MouseEvent e)
      {  
         m_XDifference = e.getX();
         m_YDifference = e.getY();
         m_dragging = true;
      }

      /**
       * End of dragging
       */             
      public void mouseReleased(MouseEvent e)
      {
         m_dragging = false;
      }
   }
}
