package src.desktop;

import src.VisualXML;
import src.gui.*;

import javax.swing.*;
import java.awt.event.*;
import java.awt.*;

import org.dom4j.*;
import java.util.Iterator;
import java.util.List;
import java.util.Collections;
import java.util.ListIterator;
import java.util.ArrayList;
import java.util.SortedMap;
import java.util.TreeMap;

/**
 * JInternalFrame for displaying content of DOMtree element. InnerFrame is able 
 * to display content of single element or a list of elements (+ some navigation)
 * Each instance can have parent InnerFrame to build some hierarchical tree-like
 * stucture and can draw a line to parent.  
 */   
public class InnerFrame extends JInternalFrame implements ActionListener, FocusListener, ItemListener
{
   private static int NODE = 0;
   private static int LIST = 1;
   private static int ATTR = 2;

   private VisualXML parent;           // main parent window
   private Element root;               // element to display
   protected Line line;                // line to the parent element-frame
   public InnerFrame pIFrame = null;   // parent element-frame
   private List<SortedMap> rowsGlobal; // all rows on the frame
   private List<InnerPanel> panels;    // list of all panels
   
   // LIST
   public boolean isList = false;          // is this element part of a list?
   protected boolean displayAsList = true; // if list, display as list [by default]
   protected int page = 1;           // current page number
   protected int pageLimit = 5;      // how many list-elements to display on a page 
   protected int listLength = 1;     // length of list (count of elements)
   
   // list of button actions - to open child element in a new instance of this class
   private SortedMap<String,Element> actionList;
   protected JPopupMenu labelPopupMenu = null;
   protected JPopupMenu listLabelPopupMenu = null;
   protected JPopupMenu attrLabelPopupMenu = null;
   protected JPopupMenu panelPopupMenu = null;
   
   protected JPopupMenu filterPopupMenu = null;
   protected SortedMap<String, Boolean> filterItems = null;
   
   protected Element elementClicked = null;
   protected Attribute attributeClicked = null;

   /**
    * Constructor
    * 
    * @param pFrame parent main window
    * @param r element to display
    * @param title title of this frame
    * @param parentIFrame parent IFrame of this IFrame                    
    */       
   public InnerFrame(VisualXML pFrame, Element r, String title, InnerFrame parentIFrame, boolean asList)
   {
      super(title, 
            false, //resizable
            true,  //closable
            false, //maximizable
            false);//iconifiable

      if (r == null) return;

      parent = pFrame; // set parent main window
      root = r;                // set root element
      pIFrame = parentIFrame;  // set parent IFrame
      displayAsList = asList;  // display as list?
      // Create popups
      labelPopupMenu = createLabelPopupMenu(false);
      listLabelPopupMenu = createLabelPopupMenu(true);
      attrLabelPopupMenu = createAttrPopupMenu();
      
      // list to display?
      Element parentNode = root.getParent();
      if (parentNode != null)
      {// has parent node
         if (parentNode.elements(root.getName()).size() > 1)
         {// is list!
            isList = true;
         }
      }
      
      filterPopupMenu = new JPopupMenu();
      // Create FilterPopup Menu
      if (isList)
      {
         // Add panel for element count filtering
         JPanel jp = new JPanel();
         jp.add(new JLabel(getString("InnerFrame.filterPopup_elementCount")));
         JTextField jtf = new JTextField("",3);
         jtf.setText("5");
         jtf.addFocusListener(
            new FocusAdapter()
            {
               public void focusLost(FocusEvent e)
               {
                  pageLimit = Integer.parseInt(((JTextField)e.getComponent()).getText());
                  page = 1;
                  updateContent();
               }
            }
         );
         jp.add(jtf);
         filterPopupMenu.add(jp);
         filterPopupMenu.addSeparator();
      }
      
      filterItems = new TreeMap<String, Boolean>();
      panelPopupMenu = createPanelPopupMenu();

      rowsGlobal = new ArrayList<SortedMap>();
      panels = new ArrayList<InnerPanel>();
    	putClientProperty("JInternalFrame.isPalette", Boolean.TRUE); // palette
      
      // Set location [sets location under the bottom most sibling or to the right of the parent]
      if (pIFrame != null)
      {// has parent IFrame
         JInternalFrame[] allFrames = parent.desktop.getAllFrames();
         // default 
         int x = pIFrame.getX() + pIFrame.getWidth() + 10;
         int y = pIFrame.getY();
         for (int i = 0; i < allFrames.length; i++)
         {// all iframes
            if (((InnerFrame)allFrames[i]).pIFrame == this.pIFrame)
            {// sibling found
               int top = allFrames[i].getY() + allFrames[i].getHeight()+10;
               if (y < top) y = top; 
            }
         }
         setLocation(x, y);
      }
      else
      {// default
         setLocation(10,10);
      }
      
      actionList = new TreeMap<String,Element>();
      if (root != null)
      {// generate content
         generateGUI();
         pack();
         drawLine();
      }
   }

   /** Sets displayAsList */
   public void setDisplayList(boolean list)
   {
      displayAsList = list;
   }

   /**
    * Sets parent element-frame
    * 
    * @param p parent element-frame [InnerFrame]        
    */       
   public void setParentInnerFrame(InnerFrame p)
   {
      pIFrame = p;
   }
   
   /**
    * Returns parent element-frame
    * 
    * @return pIFrame parent element-frame [InnerFrame]        
    */    
   public InnerFrame getParentInnerFrame()
   {
      return pIFrame;
   }
  
   /**
    * Returns root element of this JIFrame
    * 
    * @return root        
    */       
   public Element getRootElement()
   {
      return root;
   }

   /**
    * Returns a message for current localization bundle
    */       
   protected String getString(String key)
   {
      return parent.getString(key);
   }
   
   /**
    * Generates content of this IFrame. If the root element part of a list, 
    * displays each root element in its own panel.   
    */       
   protected void generateGUI()
   {
      // initialize
      rowsGlobal = new ArrayList<SortedMap>();
      panels = new ArrayList<InnerPanel>();
   
      // SET JINTERNALFRAME's LAYOUT
      GroupLayout jInternalFrame1Layout = new GroupLayout(getContentPane());
      getContentPane().setLayout(jInternalFrame1Layout);
      // HORIZONTAL  
      GroupLayout.ParallelGroup pGroup = jInternalFrame1Layout.createParallelGroup(GroupLayout.Alignment.LEADING);
      // VERTICAL
      GroupLayout.ParallelGroup vGroup = jInternalFrame1Layout.createParallelGroup(GroupLayout.Alignment.LEADING);
      GroupLayout.SequentialGroup v1Group = jInternalFrame1Layout.createSequentialGroup();        
      
      // DISPLAY AS LIST? 
      Iterator i = null;
      JLabel navigLabel = null;
      if (isList && displayAsList)
      {// it is a list -> get iterator of all elements
         i = root.getParent().elementIterator(root.getName());
         
         // add LIST NAVIGATION to the top of the JIFrame
         JPanel listNavig = new JPanel();
         listNavig.setLayout(new FlowLayout());
         String[] iconsL = {"images/left.png","images/left_p.png","images/left.png","images/left_d.png"};
         JButton left = createButton("listLeft",iconsL, getString("InnerFrame.listLeft_descr"));
         String[] iconsR = {"images/right.png","images/right_p.png","images/right.png","images/right_d.png"};
         JButton right = createButton("listRight",iconsR, getString("InnerFrame.listRight_descr"));
         navigLabel = new JLabel(page + " "+getString("InnerFrame.listCenter_of") +" "+ Math.round(listLength / pageLimit));
	       listNavig.add(left);
	       listNavig.add(navigLabel);
         listNavig.add(right);
         
         String[] iconsF = {"images/filter.png","images/filter_p.png","images/filter_mo.png","images/filter_d.png"};
         JButton filter = createButton("showFilter",iconsF, "Filter");
 
         // Add mouse listener to Filter for showing popupmenu
         filter.addMouseListener(new MouseAdapter()
         {
            public void mouseClicked(MouseEvent e) {
                filterPopupMenu.show(e.getComponent(),e.getX(), e.getY());
            }
         });
         listNavig.add(filter);
         // ADD to layout
         pGroup.addGroup(jInternalFrame1Layout.createSequentialGroup()
           .addComponent(listNavig, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE));
         v1Group.addComponent(listNavig, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE);
      }
      else
      {// only this element
         i = (Collections.singletonList((Element)root)).iterator();
      }
      
      Point range = new Point((page-1) * pageLimit, (page-1) * pageLimit + pageLimit);
      int counter = 0;
      listLength = 0;
      while(i.hasNext())
      {// GOES THROUGH SIBLINGS IN A LIST
         Element node = (Element)i.next();
         listLength++; 
         if (range.x <= counter && range.y > counter)
         {
            InnerPanel panel = createPanel(node, counter);
            panels.add(panel);
            // SET LAYOUT OF JIFRAME
            GroupLayout.SequentialGroup p1Goup = jInternalFrame1Layout.createSequentialGroup()
            .addComponent(panel, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE);
            
            v1Group.addComponent(panel, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE);
            pGroup.addGroup(p1Goup);
         }
         counter++;
      } // while ()
      
      if (navigLabel != null)
      {// change estimate pages count
         int maxPages = (listLength % pageLimit == 0) ? (int)Math.ceil(listLength / pageLimit) : ((int)Math.ceil(listLength / pageLimit)+1);
         navigLabel.setText(page  + " "+getString("InnerFrame.listCenter_of") +" "+  maxPages);
      }
      
      // FINISH THE LAYOUT OF JIFRAME
      vGroup.addGroup(v1Group);
      jInternalFrame1Layout.setHorizontalGroup(pGroup);
      jInternalFrame1Layout.setVerticalGroup(vGroup);
   }
   
   /**
    * Creates popupmenu for labels - elements and lists
    * 
    * @return new instance of JPopupMenu        
    */  
   protected JPopupMenu createLabelPopupMenu(boolean list)
   {
      // define the popup
      JPopupMenu popup = new JPopupMenu();
      popup.setBorder(BorderFactory.createLineBorder(Color.darkGray));
      JMenuItem mi = null;
      mi = new JMenuItem(parent.getString("InnerFrame.popupMenu_NodeMenuTitle"));
      mi.setEnabled(false);
      mi.setBackground(Color.darkGray);
      popup.add(mi);

      if (!list) mi = new JMenuItem(parent.getString("XMLTree.popupMenu_cloneNode"));
         else mi = new JMenuItem(parent.getString("InnerFrame.popupMenu_cloneFirstNode"));
      mi.addActionListener(this);
      mi.setActionCommand("clone");
      popup.add(mi);
    
      if (!list) mi = new JMenuItem(parent.getString("InnerFrame.popupMenu_renameNode"));
         else mi = new JMenuItem(parent.getString("InnerFrame.popupMenu_renameList"));
      mi.addActionListener(this);
      mi.setActionCommand("rename");
      popup.add(mi); 
    
      if (!list)
      {
         mi = new JMenuItem(parent.getString("XMLTree.popupMenu_insertAttr"));
         mi.addActionListener(this);
         mi.setActionCommand("insertAttr");
         popup.add(mi);
      }
      popup.addSeparator();  
      if (!list) mi = new JMenuItem(parent.getString("XMLTree.popupMenu_remove"));
         else mi = new JMenuItem(parent.getString("InnerFrame.popupMenu_removeList"));
      mi.addActionListener(this);
      mi.setActionCommand("remove");
      popup.add(mi);
      popup.setOpaque(true);
      popup.setLightWeightPopupEnabled(true);   
      return popup;
   }
   
   /**
    * Creates popupmenu for panels
    * 
    * @return new instance of JPopupMenu        
    */  
   protected JPopupMenu createPanelPopupMenu()
   {
      // define the popup
      JPopupMenu popup = new JPopupMenu();
      popup.setBorder(BorderFactory.createLineBorder(Color.darkGray));
      JMenuItem mi = null;

      // TITLE-like item
      mi = new JMenuItem("Parent menu");
      mi.setEnabled(false);
      mi.setBackground(Color.darkGray);
      popup.add(mi);

      mi = new JMenuItem(parent.getString("InnerFrame.popupMenu_insertNode"));
      mi.addActionListener(this);
      mi.setActionCommand("insertNode");
      popup.add(mi);
      mi = new JMenuItem(parent.getString("InnerFrame.popupMenu_insertAttr"));
      mi.addActionListener(this);
      mi.setActionCommand("insertAttr");
      popup.add(mi);
    
      popup.addSeparator();  
      mi = new JMenuItem(parent.getString("XMLTree.popupMenu_remove"));
      mi.addActionListener(this);
      mi.setActionCommand("removeNode");
      popup.add(mi);
    
      popup.setOpaque(true);
      popup.setLightWeightPopupEnabled(true);   
      return popup;
   }
 
   /**
    * Creates popupmenu for attributes
    * 
    * @return new instance of JPopupMenu        
    */       
   protected JPopupMenu createAttrPopupMenu()
   {
      // define the popup
      JPopupMenu popup = new JPopupMenu();
      popup.setBorder(BorderFactory.createLineBorder(Color.darkGray));
      JMenuItem mi = null;

      mi = new JMenuItem(parent.getString("InnerFrame.popupMenu_AttrTitle"));
      mi.setEnabled(false);
      mi.setBackground(Color.darkGray);
      popup.add(mi);

      mi = new JMenuItem(parent.getString("InnerFrame.popupMenu_renameAttr"));
      mi.addActionListener(this);
      mi.setActionCommand("renameAttr");
      popup.add(mi);
      popup.addSeparator();
      mi = new JMenuItem(parent.getString("InnerFrame.popupMenu_removeAttr"));
      mi.addActionListener(this);
      mi.setActionCommand("removeAttr");
      popup.add(mi);
      popup.setOpaque(true);
      popup.setLightWeightPopupEnabled(true);   
      return popup;
   }
   
   /**
    * Creates new button with action, some icons and description. 
    * Returned button has no border, its focus is not painted and it hasn't its 
    * content area filled. 
    * 
    * @param action button's action
    * @param icons array of paths to icons in this order: button, pressed, rollover, disabled
    * @param description button's description
    * @return button returns new instance of JButton     
    */       
   protected JButton createButton(String action, String[] icons, String description)
   {
      JButton button = new JButton(new ImageIcon(icons[0], description));//"images/plus.gif"
	    button.setPressedIcon(new ImageIcon(icons[1], description));//"images/plus_p.gif"
	    button.setRolloverIcon(new ImageIcon(icons[2], description));//"images/plus_r.gif"
	    button.setDisabledIcon(new ImageIcon(icons[3], description));
	    if (!description.equals("")) button.setToolTipText(description);
	    ((AbstractButton)button).setBorderPainted(false);
	    ((AbstractButton)button).setFocusPainted(false);
	    ((AbstractButton)button).setContentAreaFilled(false);
      button.setMargin(new Insets(0,0,0,0));
      button.setActionCommand(action);
      button.addActionListener(this);
          
      return button;
   }

   /**
    * Function for handles changes in filter popupmenu. 
    * [Filters displayed elements in a list]
    */       
   public void itemStateChanged(ItemEvent e)
   {
      JMenuItem source = (JMenuItem)(e.getSource());
      // insert new element to filter
      filterItems.put(source.getText(),new Boolean((e.getStateChange() == ItemEvent.SELECTED)));
        
      Iterator i = filterItems.keySet().iterator();
      while (i.hasNext())
      {
         String key = (String) i.next();
      }
      updateContent();
   }

   /**
    * For showing popupmenus
    */       
   class LabelMouseAdapter extends MouseAdapter
   {
      public void mouseClicked(MouseEvent e) {
      if(e.getButton() == MouseEvent.BUTTON3) {
         // find target element 
         Iterator i = rowsGlobal.iterator();
         while(i.hasNext())
         {
            SortedMap map = (SortedMap) i.next();
            Iterator ii = map.values().iterator();
            while (ii.hasNext())
            {
               Row r = (Row) ii.next();
               if (r.label == e.getComponent())
               {
                  if (r.isNode())
                  {
                     labelPopupMenu.show(e.getComponent(), e.getX(), e.getY());
                  }
                  else if (r.isList())
                  {
                     listLabelPopupMenu.show(e.getComponent(), e.getX(), e.getY());
                  }
                  else if (r.isAttribute())
                  {
                     attrLabelPopupMenu.show(e.getComponent(), e.getX(), e.getY());
                  }
                  elementClicked = r.e;
                  attributeClicked = r.a;
               }
            }
	}
     }
   }
   }

   /**
    * Creates panel with all content of the node. Text-only child elements will 
    * be displayed with an "text icon" and a textfield. If child element is a 
    * branch (has child elements) or is a part of a list, it will be displayed 
    * with a "plus" button [which opens new IFrame]. Elements with attributes 
    * are displayed with names and values of these attributes attached to the 
    * name of the element.
    * 
    * @param node root element to display
    * @param c sequence number of this root element
    * @return panel returns new instance of panel with its content                
    */       
   protected InnerPanel createPanel(Element node, int c)
   {
      // CREATE NEW PANEL AND SET IT's LAYOUT
      InnerPanel panel = new InnerPanel(node);
        
      panel.addMouseListener(new MouseAdapter() {
            public void mouseClicked(MouseEvent e) {
                if(e.getButton() == MouseEvent.BUTTON3)
                {
                    Iterator i = panels.iterator();
         while(i.hasNext())
         {
            InnerPanel p = (InnerPanel) i.next();
            if (p == e.getComponent())
            {
               elementClicked = p.getRootElement();
              }
            }
                   panelPopupMenu.show(e.getComponent(),e.getX(), e.getY());
                   
                }
            }
        });
      
      panel.setBorder(BorderFactory.createTitledBorder(getTitle()));
      GroupLayout panelLayout = new GroupLayout(panel);
      panel.setLayout(panelLayout);
      // HORIZONTAL GROUP
      GroupLayout.ParallelGroup h = panelLayout.createParallelGroup(GroupLayout.Alignment.LEADING);
      GroupLayout.SequentialGroup h1 = panelLayout.createSequentialGroup();
      GroupLayout.ParallelGroup buttons = panelLayout.createParallelGroup(GroupLayout.Alignment.TRAILING, false);     
      GroupLayout.ParallelGroup texts = panelLayout.createParallelGroup(GroupLayout.Alignment.LEADING);
         
      // VERTICAL GROUP
      GroupLayout.ParallelGroup v = panelLayout.createParallelGroup(GroupLayout.Alignment.LEADING);
      GroupLayout.SequentialGroup v1 = panelLayout.createSequentialGroup();

      //----------------------------------------------------- LIST OF ATTRIBUTES
      SortedMap<String,Row> rows = new TreeMap<String,Row>();
      List attributes = node.attributes();
      Iterator attrIterator = attributes.iterator();
      int i = 0;
      while(attrIterator.hasNext())
      {// GOES THROUGH ALL ATTRIBUTES
         Attribute attr = (Attribute)attrIterator.next();

         // INSERT INTO FILTER
         if (!filterItems.containsKey("attr:"+attr.getName()))
         {
            filterItems.put("attr:"+attr.getName(), new Boolean(true));
            JCheckBoxMenuItem check = new JCheckBoxMenuItem("attr:"+attr.getName(), true);
            check.addItemListener(this);
            filterPopupMenu.add(check);
         }
         else
         {
            if (!filterItems.get("attr:"+attr.getName()).booleanValue())
            {
               continue;
            }
         }

          
         // CREATE CONTENT...
         JLabel icon = new JLabel(new ImageIcon("./images/attr.gif"));
         JLabel name = new JLabel("attr:"+attr.getName());
         name.addMouseListener(new LabelMouseAdapter());
         
         JText value = new JText(attr.getValue(), node, attr);
         value.addFocusListener(this);
         
         // STORE ROW (ELEMENT) FOR LATER CHECKING
         Row r = new Row(node, attr, icon, name, value);
         r.setIsAttribute();
         rows.put("attr:"+attr.getName()+":"+c,r);
         // SET THE LAYOUT OF PANEL        
         // HORIZONTAL
         GroupLayout.SequentialGroup labels = panelLayout.createSequentialGroup();  
         labels.addComponent(icon);
         labels.addGap(5);
         labels.addComponent(name);
         labels.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED, 10, Short.MAX_VALUE);
         labels.addComponent(value, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE);
         texts.addGroup(labels);

         // VERTICAL
         GroupLayout.ParallelGroup row = panelLayout.createParallelGroup(GroupLayout.Alignment.BASELINE);   
         row.addComponent(icon);
         row.addComponent(name);
         row.addComponent(value, GroupLayout.PREFERRED_SIZE, 20, GroupLayout.PREFERRED_SIZE);
         v1.addGroup(row);
         v.addGroup(v1);
      }// while ()

      //---------------------------------------------------- LIST OF ELEMENTS
      List elements = node.elements();
      Iterator elementIterator = elements.iterator();
       
      

      while(elementIterator.hasNext())
      {// GOES THROUGH ALL CHILD ELEMENTS
         Element element = (Element)elementIterator.next();
         
         // INSERT INTO FILTER
         if (!filterItems.containsKey(element.getName()))
         {
            filterItems.put(element.getName(), new Boolean(true));
            JCheckBoxMenuItem check = new JCheckBoxMenuItem(element.getName(), true);
            check.addItemListener(this);
            filterPopupMenu.add(check);
         }
         else
         {
            if (!filterItems.get(element.getName()).booleanValue())
            {
               continue;
            }
         }
         
         
         // CHECK WHETHER THERE's BEEN SUCH ELEMENT BEFORE 
         if (rows.containsKey(element.getName()))
         {// IS PART OF A LIST
            Row r = rows.get(element.getName());
            //Row rGlobal = rowsGlobal.get(element);
            // change icon
            r.icon.setIcon(new ImageIcon("images/list.gif"));
            // increase list size
            r.count++;
            r.button.setEnabled(true);
            r.text.setVisible(false);
            // change name
            r.label.setText(element.getName()+" ["+r.count+"]");
            r.setIsList();
            //rGlobal.setIsList();
            continue;
         }
          
         // ACTION FOR BUTTON (TO SHOW CHILD ELEMENTS)
         String action = "show<"+element.getName()+"-"+c+">";
         actionList.put(action, element);
          
         // CREATE CONTENT...
         String icons[] = {"images/plus.gif","images/plus_p.gif","images/plus_r.gif","images/plus_d.gif"};
         JButton button = createButton(action, icons, getString("InnerFrame.buttonShow_descr"));//new JButton();
         JLabel icon = new JLabel();
         JLabel name = new JLabel(element.getName());
         
         name.addMouseListener(new LabelMouseAdapter());
         
         
         JText value = new JText("",null); 

         boolean editable = false;
         // IF ANY ATTRIBUTES -> PRINT THEM TOO 
         if (element.attributeCount()>0)
         {
            Iterator attrI = element.attributeIterator();
            String str = element.getName() + " ";
            while (attrI.hasNext())
            {// attach to the end of element's name
               Attribute attr = (Attribute) attrI.next();
               str = str+" "+attr.getName()+"=\""+attr.getValue()+"\" ";
            }
            name.setText(str);
         }

         // TEST: ELEMENT HAS ONLY TEXT VALUE AND NO ATTRs
         if (element.isTextOnly() && element.attributeCount()==0) 
         {// yes
            button.setEnabled(false);
            // set icon
            icon.setIcon(new ImageIcon("./images/text.gif"));
            value = new JText(element.getText(), element);
            value.addFocusListener(this);
            // mark as editable
            editable = true;
         }
         else
         {// no
            icon.setIcon(new ImageIcon("./images/record.gif"));
         }
          
         // STORE ROW (ELEMENT) FOR LATER CHECKING
         rows.put(element.getName(),new Row(element,icon,button, name, value));
          
         // SET THE LAYOUT OF PANEL        
         // HORIZONTAL
         GroupLayout.SequentialGroup labels = panelLayout.createSequentialGroup();  
         buttons.addComponent(button, GroupLayout.Alignment.LEADING, GroupLayout.PREFERRED_SIZE, 14, Short.MAX_VALUE);
         labels.addComponent(icon);
         labels.addGap(5);
         labels.addComponent(name);
         labels.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED, 10, Short.MAX_VALUE);
         if (editable) labels.addComponent(value, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE);
         texts.addGroup(labels);

         // VERTICAL
         GroupLayout.ParallelGroup row = panelLayout.createParallelGroup(GroupLayout.Alignment.BASELINE);   
         row.addComponent(button);//if (!editable) 
         row.addComponent(icon);
         row.addComponent(name);
         if (editable) row.addComponent(value, GroupLayout.PREFERRED_SIZE, 20, GroupLayout.PREFERRED_SIZE);
         v1.addGroup(row);
         v.addGroup(v1);
      }// while ()
         
      //---------------------------------------------------------- TEXT VALUE
      if (node.isTextOnly() || node.getTextTrim().length() > 0)
      {
         JLabel icon = new JLabel(new ImageIcon("./images/text.gif"));
         JText value =  new JText(node.getTextTrim(),node);
         value.addFocusListener(this);
         JLabel name =  new JLabel("this");

         // HORIZONTAL
         GroupLayout.SequentialGroup labels = panelLayout.createSequentialGroup();  
         labels.addComponent(icon);
         labels.addGap(5);
         labels.addComponent(name);
         labels.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED, 10, Short.MAX_VALUE);
         labels.addComponent(value, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE);
         texts.addGroup(labels);

         // VERTICAL
         GroupLayout.ParallelGroup row = panelLayout.createParallelGroup(GroupLayout.Alignment.BASELINE);   
         row.addComponent(icon);
         row.addComponent(name);
         row.addComponent(value, GroupLayout.PREFERRED_SIZE, 20, GroupLayout.PREFERRED_SIZE);
         v1.addGroup(row);
         v.addGroup(v1);
      }
         
      // FINISH PANEL LAYOUT
      h1.addGroup(buttons);
      h1.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED);
      h1.addGroup(texts);
      //h1.addContainerGap();
      h.addGroup(h1);
      panelLayout.setHorizontalGroup(h);
      panelLayout.setVerticalGroup(v);
      
      rowsGlobal.add(rows);
      return panel;
   }

   /**
    * Class for storing pointers on components in one row of panel 
    */       
	 protected class Row
	 {
 
      public Element e = null;
      public Attribute a = null;
      public JLabel icon;
      public JButton button;
      public JLabel label;
      public JTextField text;
      public int count;
      public int type = NODE; // type of node
      //public Type nodeType;
     
      public Row(Element elem, JLabel ico, JButton butt, JLabel lab, JTextField txt)
      {
         e = elem;
         icon = ico;
         button = butt;
         label = lab;
         text = txt;
         count = 1;
      }
      
      public Row(Element elem, Attribute attr, JLabel ico, JLabel lab, JTextField txt)
      {
         e = elem;
         a = attr;
         icon = ico;
         button = null;
         label = lab;
         text = txt;
         count = 1;
      }
      
      public void setIsList()
      {
         type = LIST;
      }
      public void setIsAttribute()
      {
         type = ATTR;
      }
      public void setIsNode()
      {
         type = NODE;
      }
      public boolean isList()
      {
         return (type == LIST);
      }
      public boolean isAttribute()
      {
         return (type == ATTR);
      }
      public boolean isNode()
      {
         return (type == NODE);
      }      
   }

   
   /**
    * Called when one of the fields gets the focus. Needed for FocusListener 
    * interface.
    */
   public void focusGained(FocusEvent e)
   {
   }
   
   /**
    * Called when one of the fields losts the focus. Stores changed text into 
    * corresponding DOM element. Changes the state of document to "Changed"
    * Needed for FocusListener interface.
    */       
   public void focusLost(FocusEvent e)
   {
      Component c = e.getComponent();
      boolean changed = false;
      
      if (((JText)c).getAttribute()!=null)
      {// Attribute edited -> SAVE
         if (!((JText)c).getAttribute().getValue().equals(((JText)c).getText()))
         {
            ((JText)c).getAttribute().setValue(((JText)c).getText());
            changed = true;
            pIFrame.updateContent();
         }
      }
      else if (((JText)c).getElement()!=null)
      {// text element edited -> SAVEx
         if (!((JText)c).getElement().getText().equals(((JText)c).getText()))
         {
            ((JText)c).getElement().setText(((JText)c).getText());
            changed = true;
         }
      }
      else
      {
         return;
      }
       
      if (changed && !parent.getEdited())
      {
         parent.setEdited(true);
      }
   }

   /**
    * Gets parent element-frame (used for creating tree-like structure)
    */       
   public InnerFrame getParentFrame()
   {
      return pIFrame;
   }
   
   /**
    * Ordinary JTextField with pointer to corresponding DOM element
    */       
   class JText extends JTextField
   {
      protected Element e;
      protected Attribute a;
      
      public JText(String str, Element elem)
      {
         super(str, 10);
         e = elem;
         a = null;
      }

      public JText(String str, Element elem, Attribute attr)
      {
         super(str, 10);
         e = elem;
         a = attr;
      }
      
      public Element getElement()
      {
         return e;
      }
      public Attribute getAttribute()
      {
         return a;
      }
   }

   /**
    * Class contains two Points (starting and ending).
    */          
   class Vector2D
   {
      Point p1;
      Point p2;
      double length;
      
      /** Constructor */
      Vector2D(Point P1, Point P2)
      {
         p1 = P1;
         p2 = P2;
         length = Math.sqrt(Math.pow(p1.x-p2.x, 2)+Math.pow(p1.y-p2.y, 2));
      }
      
      /** Returns length of the vector */
      public double getLength()
      {
         return length;
      }
      
      /** Returns first point */
      public Point getStartPoint()
      {
         return p1;
      }

      /** Returns second point */   
      public Point getEndPoint()
      {
         return p2;
      }
      
      /** Returns tan of the vector in degrees */
      public double getAngleInDegrees()
      {
         double tan = (double)(p2.y-p1.y)/(p2.x-p1.x);
         return Math.toDegrees(Math.atan(tan));
      }

      /** Returns tan of the vector */      
      public double getAngle()
      {
         return Math.atan((double)(p2.y-p1.y)/(p2.x-p1.x));
      }
   }
   
   /**
    * Function draws a line to the parent iframe. 
    */       
   public void drawLine()
   {
      // no parent JIFrame.. 
      if (pIFrame == null)
      {
         return;
      }
    
      Point p = null;
      Point c = null;
      if (pIFrame.isList)
      {// stick the line to "parent panel"
         Rectangle bounds = pIFrame.getPanelLocation(root.getParent());
         p = new Point(pIFrame.getX()+bounds.x+bounds.width/2, bounds.y+pIFrame.getY()+bounds.height/2+15);
         c = new Point(this.getX()+this.getWidth()/2, this.getY()+this.getHeight()/2);
      }
      else
      {
         // positions of both frames (parent, child)
         p = new Point(pIFrame.getX()+pIFrame.getWidth()/2, pIFrame.getY()+pIFrame.getHeight()/2);
         c = new Point(this.getX()+this.getWidth()/2, this.getY()+this.getHeight()/2);
      }   
      Point p1 = null;
      Point p2 = null;
      Point bounds = null;
         
      if (p.x<=c.x && p.y<=c.y)
      {// IV
         Point pB = new Point(p.x, pIFrame.getY()+pIFrame.getHeight());
         Point pR = new Point(pIFrame.getX()+pIFrame.getWidth(), p.y);
         Point chL = new Point(getX(), c.y);
         Point chT = new Point(c.x, getY());
            
         Vector2D lines[] = new Vector2D[4];
         lines[0] = new Vector2D(pB, chL);
         lines[1] = new Vector2D(pB, chT);
         lines[2] = new Vector2D(pR, chL);
         lines[3] = new Vector2D(pR, chT);
            
         // check visibility of the new line
         int index; // index of line
         for(index = 0; index < lines.length-1; index++)
         {// find the most appropriate line
            double angle = lines[index].getAngleInDegrees();
            if (angle <= 90 && angle >= 0) break;
         }
            
         p1 = lines[index].getStartPoint();
         p2 = lines[index].getEndPoint();
         bounds = new Point(p2.x, p2.y);
      }
      else if (p.x<c.x && p.y>c.y)
      {// I
         Point pT = new Point(p.x, pIFrame.getY());
         Point pR = new Point(pIFrame.getX()+pIFrame.getWidth(), p.y);
         Point chL = new Point(getX(), c.y);
         Point chB = new Point(c.x, getY()+getHeight());
            
         Vector2D lines[] = new Vector2D[4];
         lines[0] = new Vector2D(pR, chL);
         lines[1] = new Vector2D(pT, chL);
         lines[2] = new Vector2D(pT, chB);
         lines[3] = new Vector2D(pR, chB);
            
         // check visibility of the new line
         int index;
         for(index = 0; index < lines.length-1; index++)
         {// find the most appropriate line
            double angle = lines[index].getAngleInDegrees();
            if (angle <= 0) break;
         }
            
         p1 = lines[index].getStartPoint();
         p2 = lines[index].getEndPoint();
         bounds = new Point(p2.x, p1.y);
      }
      else if (p.x>c.x && p.y>c.y)
      {// II
         Point pT = new Point(p.x, pIFrame.getY());
         Point pL = new Point(pIFrame.getX(), p.y);
         Point chR = new Point(getX()+getWidth(), c.y);
         Point chB = new Point(c.x, getY()+getHeight());

         Vector2D lines[] = new Vector2D[4];
         lines[0] = new Vector2D(pL, chR);
         lines[1] = new Vector2D(pT, chB);
         lines[2] = new Vector2D(pT, chR);
         lines[3] = new Vector2D(pL, chB);
            
         // check visibility of the new line
         int index;
         for(index = 0; index < lines.length-1; index++)
         {// find the most appropriate line
            double angle = lines[index].getAngleInDegrees();
            if (angle <= 90 && angle >= 0) break;
         }
            
         p1 = lines[index].getStartPoint();
         p2 = lines[index].getEndPoint();
         bounds = new Point(p1.x, p1.y);
      }
      else
      {// III
         Point pB = new Point(p.x, pIFrame.getY()+pIFrame.getHeight());
         Point pL = new Point(pIFrame.getX(), p.y);
         Point chR = new Point(getX()+getWidth(), c.y);
         Point chT = new Point(c.x, getY());

         Vector2D lines[] = new Vector2D[4];
         lines[0] = new Vector2D(pB, chT);
         lines[1] = new Vector2D(pL, chR);
         lines[2] = new Vector2D(pB, chR);
         lines[3] = new Vector2D(pL, chT);
            
         // check visibility of the new line
         int index;
         for(index = 0; index < lines.length-1; index++)
         {// find the most appropriate line
            double angle = lines[index].getAngleInDegrees();
            if (angle <= 0) break;
         }
            
         p1 = lines[index].getStartPoint();
         p2 = lines[index].getEndPoint();
         bounds = new Point(p1.x, p2.y);
      }
         
      if(p1 == null || p2 == null)
      {
         return;
      }
         
      if (this.line != null)
      {// redraw existing component (line)
         line.setBounds(0,0,bounds.x+5, bounds.y+5);
         line.redraw(p1.x, p1.y, p2.x, p2.y);
      }
      else
      {// create new one
         Line l = new Line(p1.x, p1.y, p2.x, p2.y);
         l.setBounds(0,0,bounds.x+5, bounds.y+5);
         parent.desktop.add(l);
         this.line = l;
      }
   }
    
   /**
    * Handles actions from the actionList(buttons) and from listNavigation
    */       
   public void actionPerformed(ActionEvent e)
   {
      // Actions for plus buttons to show child elements
      for (String action : actionList.keySet())
      {
         if (action.equals(e.getActionCommand()))
         {// Action command found
            InnerFrame newFrame = parent.createFrame(actionList.get(action), this.getTitle()+"/"+actionList.get(action).getName(), this);
            /*if (isList)
            {
               Rectangle bounds = getPanelLocation(actionList.get(action).getParent());
               if (bounds != null)
               {
                  newFrame.setLocation(bounds.x+bounds.width+10, bounds.y);
               }
            }*/
         }
      }
      
      // Button in list navigation
      if ("listLeft".equals(e.getActionCommand()))
      {// Left
         if (page > 1)
         {
            killChildIFrames(this);
            page--;
            updateContent();
         }
      }
      else if ("listRight".equals(e.getActionCommand()))
      {// Right
         if (page*pageLimit < listLength)
         {
            killChildIFrames(this);
            page++;
            updateContent();
         }
      }
      else if ("insertAttr".equals(e.getActionCommand()))
      {// Insert attribute to an element
         String s = (String)JOptionPane.showInputDialog(parent,
                            ((VisualXML)parent).getString("InnerFrame.dialog_enterAttrName"),
                            ((VisualXML)parent).getString("InnerFrame.dialog_enterAttrName_title"),
                            JOptionPane.PLAIN_MESSAGE,null,null,"");

         // String returned 
         if ((s != null) && (s.length() > 0) && (elementClicked.attribute(s) == null))
         {
            elementClicked.addAttribute(s,"");   // insert attr
            parent.updateIFrames(null);          // update content of all iframes
            ((VisualXML)parent).setEdited(true); // set document changed
      
            // Update XMLTree
            ExtendedTreeNode etn = parent.tree.findNode(elementClicked);
            if (etn != null) parent.tree.refreshNode(etn);
         }
      }
      else if ("clone".equals(e.getActionCommand()))
      {// deep clone of an element
         ExtendedTreeNode etn = parent.tree.findNode(elementClicked);
         if (etn != null)
         {// clone node in tree
            parent.tree.cloneNode(etn);
            parent.updateIFrames(null);
            ((VisualXML)parent).setEdited(true);
         }
      }
      else if ("remove".equals(e.getActionCommand()))
      {// removes all elements of given name.. (removes whole list too)
         // find target element 
         Iterator i = elementClicked.getParent().elementIterator(elementClicked.getName());
         Element parentElem = elementClicked.getParent();
         while (i.hasNext())
         {// remove all elements
            Element elem = (Element) i.next();
            parentElem.remove(elem);
            ExtendedTreeNode etn = parent.tree.findNode(elem);
            if (etn != null) parent.tree.removeNode(etn);
         }
         ((VisualXML)parent).setEdited(true);
         parent.updateIFrames(null);
      }
      else if ("rename".equals(e.getActionCommand()))
      {// rename
         String s = (String)JOptionPane.showInputDialog(parent,
                                ((VisualXML)parent).getString("InnerFrame.dialog_rename"),
                                ((VisualXML)parent).getString("InnerFrame.dialog_rename_title"),
                                JOptionPane.PLAIN_MESSAGE,null,null,"");
         // String returned 
         if ((s != null) && (s.length() > 0))
         {
            // find target element 
            Iterator i = elementClicked.getParent().elementIterator(elementClicked.getName());
            while (i.hasNext())
            {
               Element elem = (Element) i.next();
               elem.setName(s);
               ExtendedTreeNode etn = parent.tree.findNode(elem);
               if (etn != null) parent.tree.renameNode(etn,s);
            }
            ((VisualXML)parent).setEdited(true);
            parent.updateIFrames(null);
         }
      }
      else if ("removeAttr".equals(e.getActionCommand()))
      {// removes attribute
         // find parent element 
         Iterator i = rowsGlobal.iterator();
         while(i.hasNext())
         {
            SortedMap map = (SortedMap) i.next();
            Iterator ii = map.values().iterator();
            while (ii.hasNext())
            {
               Row r = (Row) ii.next();
               if (r.a == attributeClicked)
               {// attr and parent element found
                  r.e.remove(r.a);
                  ((VisualXML)parent).setEdited(true);
                  parent.updateIFrames(null);
                  
                  ExtendedTreeNode etn = parent.tree.findNode(r.e);
                  if (etn != null) parent.tree.refreshNode(etn);
                  return;
               }
            } 
         }
      }
      else if ("renameAttr".equals(e.getActionCommand()))
      {// renames attribute
         // find parent element 
         Iterator i = rowsGlobal.iterator();
         while(i.hasNext())
         {
            SortedMap map = (SortedMap) i.next();
            Iterator ii = map.values().iterator();
            while (ii.hasNext())
            {
               Row r = (Row) ii.next();
               if (r.a == attributeClicked)
               {
                  String s = (String)JOptionPane.showInputDialog(parent,
                                     ((VisualXML)parent).getString("InnerFrame.dialog_rename"),
                                     ((VisualXML)parent).getString("InnerFrame.dialog_rename_title"),
                                     JOptionPane.PLAIN_MESSAGE,null,null, "");

                  // String returned 
                  if ((s != null) && (s.length() > 0) && (r.e.attribute(s) == null))
                  {
                     r.e.addAttribute(s,r.a.getValue());
                     r.e.remove(r.a);
                     ((VisualXML)parent).setEdited(true);
                     parent.updateIFrames(null);
                  
                     // refresh XMLtree
                     ExtendedTreeNode etn = parent.tree.findNode(r.e);
                     if (etn != null) parent.tree.refreshNode(etn);
                     return;
                  }
               }
            } 
         }
      }
      else if ("removeNode".equals(e.getActionCommand()))
      {// removes node
         Element parentElem = elementClicked.getParent();
         if (parentElem != null)
         {
            parentElem.remove(elementClicked);
            ExtendedTreeNode etn = parent.tree.findNode(elementClicked);
            if (etn != null) parent.tree.removeNode(etn);

            if (!isList)
            {// this was the only panel on the frame -> close
               dispose();
            }
            else
            {
               if (root == elementClicked)
               {// we have to set another element from the list, root's been removed
                  Iterator i = parentElem.elementIterator(root.getName());
                  if (i.hasNext()) root = (Element) i.next();
                     else dispose();
               }
            }

            ((VisualXML)parent).setEdited(true);
            parent.updateIFrames(null);
         }
      }
      else if ("insertNode".equals(e.getActionCommand()))
      {// inserts node
         String s = (String)JOptionPane.showInputDialog(parent,
                            ((VisualXML)parent).getString("InnerFrame.dialog_enterName"),
                            ((VisualXML)parent).getString("InnerFrame.dialog_enterName_title"),
                            JOptionPane.PLAIN_MESSAGE, null,null, "");

         // String returned 
         if ((s != null) && (s.length() > 0))
         {
            Element elem = elementClicked.addElement(s);
            ExtendedTreeNode etn = parent.tree.findNode(elementClicked);
            if (etn != null) parent.tree.addNode(etn, elem);

            ((VisualXML)parent).setEdited(true);
            parent.updateIFrames(null);
         }
      }
   }
   
   /**
    * Updates content (recreates all components) and redraws Line to parent iframe
    */        
   public void updateContent()
   {
      getContentPane().removeAll();
      generateGUI();
      pack();
      drawLine();
   }

   /**
    * Removes all child iframes.
    * 
    * @param iframe parent iframe        
    */       
   public void killChildIFrames(InnerFrame iframe)
   {
      JInternalFrame allFrames[] = parent.desktop.getAllFrames();
      for (int i = 0; i < allFrames.length; i++)
      {// kill all children
         if (((InnerFrame)allFrames[i]).getParentFrame() == iframe)
         {
            parent.desktop.remove(allFrames[i]); // kill all children of this child.... :)
         }
      }
   }

   /**
    * Returns location of panel containing given element
    * 
    * @param e element
    * @return location of panel containing given element
    */       
   public Rectangle getPanelLocation(Element e)
   {
      Iterator i = panels.iterator();
      while (i.hasNext())
      {
         InnerPanel p = (InnerPanel) i.next();
         if (p.getRootElement() == e)
         {// panel found
            return p.getBounds();
         }
      }
      return null;
   }
}
