package src.desktop;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.beans.*;

/**
 * An extension of WDesktopPane that supports often used MDI functionality. This
 * class also handles setting scroll bars for when windows move too far to the left or
 * bottom, providing the MDIDesktopPane is in a ScrollPane.
 */
public class MDIDesktopPane extends JDesktopPane implements MouseMotionListener
{
    private static int FRAME_OFFSET=20;
    private MDIDesktopManager manager;

    public MDIDesktopPane()
    {
       manager=new MDIDesktopManager(this);
       setDesktopManager(manager);
       setDragMode(JDesktopPane.OUTLINE_DRAG_MODE);
        
       setAutoscrolls(true); //enable synthetic drag events
       addMouseMotionListener(this); //handle mouse drags
    }

    public void mouseDragged(MouseEvent e) {
        //The user is dragging us, so scroll!
        Rectangle r = new Rectangle(e.getX(), e.getY(), 1, 1);
        scrollRectToVisible(r);
    }
    
    public void mouseMoved(MouseEvent e)
    {
    }
    
    public void setBounds(int x, int y, int w, int h) {
        super.setBounds(x,y,w,h);
        checkDesktopSize();
    }

    public Component add(JInternalFrame frame)
    {
        Component retval=super.add(frame);
        checkDesktopSize();
        return retval;
    }

    /**
     * Sets all component size properties ( maximum, minimum, preferred)
     * to the given dimension.
     */
    public void setAllSize(Dimension d){
        setMinimumSize(d);
        setMaximumSize(d);
        setPreferredSize(d);
    }

    /**
     * Sets all component size properties ( maximum, minimum, preferred)
     * to the given width and height.
     */
    public void setAllSize(int width, int height){
        setAllSize(new Dimension(width,height));
    }

    public void remove(Component c)
    {
        if (c instanceof InnerFrame)
        {// remove line..
           if (((InnerFrame)c).line != null) super.remove(((InnerFrame)c).line);
        }
        
        // REMOVE WHOLE BRANCH
        JInternalFrame allFrames[] = getAllFrames();
        for (int i = 0; i < allFrames.length; i++)
        {// kill all children
           if (((InnerFrame)allFrames[i]).getParentFrame() == c)
           {
              super.remove(((InnerFrame)allFrames[i]).line);
              remove(allFrames[i]); // kill all children of this child.... :)
           }
        }
        
        super.remove(c);
        super.repaint();
        checkDesktopSize();
    }

    private void checkDesktopSize() {
        if (getParent()!=null&&isVisible()) manager.resizeDesktop();
    }
}

/**
 * Private class used to replace the standard DesktopManager for JDesktopPane.
 * Used to provide scrollbar functionality.
 */
class MDIDesktopManager extends DefaultDesktopManager {
    private MDIDesktopPane desktop;
    private int old_x = -1;
    private int old_y = -1;
    private int max_x = 0;
    private int max_y = 0;
    public MDIDesktopManager(MDIDesktopPane desktop) {
        this.desktop = desktop;
    }

    public void endResizingFrame(JComponent f) {
        super.endResizingFrame(f);
        resizeDesktop();
    }

    public void endDraggingFrame(JComponent f) {
        super.endDraggingFrame(f);
        resizeDesktop();
    }

    public void resizeFrame(JComponent f, int newX, int newY, int newWidth, int newHeight)
    {
        super.resizeFrame(f, newX, newY, newWidth, newHeight);
        if (f instanceof InnerFrame)
        {
           JInternalFrame allFrames[] = desktop.getAllFrames();
           for (int i = 0; i < allFrames.length; i++)
           {
              if (allFrames[i] == f)
              {
                 ((InnerFrame)allFrames[i]).drawLine();
              }
           }
        } 
    }

    public void dragFrame(JComponent f, int newX, int newY)
    {
       super.dragFrame(f, newX, newY);
       if (f instanceof InnerFrame)
       {
          ((InnerFrame)f).drawLine();
          JInternalFrame allFrames[] = desktop.getAllFrames();
          for (int i = 0; i < allFrames.length; i++) {

          // redraw lines from child JIFrames
          if (((InnerFrame)allFrames[i]).getParentFrame() == f)
          {
             ((InnerFrame)allFrames[i]).drawLine();
          }                
          }
       } 
    }

    public void setNormalSize()
    {
        JScrollPane scrollPane=getScrollPane();
        int x = 0;
        int y = 0;
        Insets scrollInsets = getScrollPaneInsets();

        if (scrollPane != null) {
            Dimension d = scrollPane.getVisibleRect().getSize();
            if (scrollPane.getBorder() != null) {
               d.setSize(d.getWidth() - scrollInsets.left - scrollInsets.right,
                         d.getHeight() - scrollInsets.top - scrollInsets.bottom);
            }

            d.setSize(d.getWidth() - 20, d.getHeight() - 20);
            desktop.setAllSize(x,y);
            scrollPane.invalidate();
            scrollPane.validate();
        }
    }

    private Insets getScrollPaneInsets() {
        JScrollPane scrollPane=getScrollPane();
        if (scrollPane==null) return new Insets(0,0,0,0);
        else return getScrollPane().getBorder().getBorderInsets(scrollPane);
    }

    private JScrollPane getScrollPane() {
        if (desktop.getParent() instanceof JViewport) {
            JViewport viewPort = (JViewport)desktop.getParent();
            if (viewPort.getParent() instanceof JScrollPane)
                return (JScrollPane)viewPort.getParent();
        }
        return null;
    }
    
    /**
     * Loads background image and tiles it on the desktop.
     */         
    protected void loadBackgroundImage(int newX, int newY)
    {
       ImageIcon icon = new ImageIcon("images/bg.png");
       if (newX > old_x || newY > old_y)
       {// only if the size increased
          int xx = old_x;
          int yy = old_y;

          for (int x=0; x < desktop.getWidth(); x += icon.getIconWidth())
          {
             for (int y=0; y < desktop.getHeight(); y += icon.getIconHeight())
             {
                 if (x > old_x || y > old_y)
                 {
                    JLabel l = new JLabel(icon);
                    l.setBounds(x,y,icon.getIconWidth(),icon.getIconHeight());
                    desktop.add(l, new Integer(Integer.MIN_VALUE));
                    yy = y;
                    xx = x;
                 }
             }
          }
          old_y = yy;
          old_x = xx;
       }
    }
    
    protected void resizeDesktop() {
        int x = 0;
        int y = 0;
        JScrollPane scrollPane = getScrollPane();
        Insets scrollInsets = getScrollPaneInsets();

        if (scrollPane != null) {
            JInternalFrame allFrames[] = desktop.getAllFrames();
            for (int i = 0; i < allFrames.length; i++) {
                if (allFrames[i].getX()+allFrames[i].getWidth()>x) {
                    x = allFrames[i].getX() + allFrames[i].getWidth();
                }
                if (allFrames[i].getY()+allFrames[i].getHeight()>y) {
                    y = allFrames[i].getY() + allFrames[i].getHeight();
                }
            }
            Dimension d=scrollPane.getVisibleRect().getSize();
            if (scrollPane.getBorder() != null) {
               d.setSize(d.getWidth() - scrollInsets.left - scrollInsets.right,
                         d.getHeight() - scrollInsets.top - scrollInsets.bottom);
            }
            
            if (x > old_x || y > old_y)
            {// new size is greater
               int newX = old_x;
               int newY = old_y;
               if (x > old_x) newX = x;
               if (y > old_y) newY = y;

               desktop.setAllSize(newX,newY);
               loadBackgroundImage(newX,newY);
               scrollPane.invalidate();
               scrollPane.validate();
            }
        }
    }
}
